/* TROIA distributor
 * version 0.1 date 2004-08-31
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <malloc.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#include "protocols.h"
#include "td_config.h"
#include "td_fmt.h"

//global variables
int end = 0; //set to 1 by signal handler to indicate end of program

//signal handler to end program
static void end_signal( int sig_no )
{
  //end program
  end = 1;
  //keep compiler happy
  sig_no = 0;
}

//convert frame data of a device into output format
//(see specification of TROIA distribution module for detailed explanation)
static void frame2out( unsigned char * * p_frame, unsigned char * p_out )
{
  static unsigned short ser_no, pix_no;
  register unsigned char byte, val;
  register unsigned char bit, byte_bits;

  //pixels 0 ... pix_cnt-1
  for( pix_no = 0; pix_no < td_out_pix_cnt; pix_no++ )
  {
    //bits D1 ... D7 (we've only got 128 greyscales)
    for( bit = 1; bit < 8; bit++ )
    {
      byte_bits = 0; //start with empty byte
      byte = 0;
      //serial ports 0 ... ser_cnt-1
      for( ser_no = 0; ser_no < td_out_ser_cnt; ser_no++ )
      {
        //this command sequence is faster than a long expression
        //if-statements and ?-:-expresions are to be avoided for speed
        //testes with gcc-2.95.4 and gcc-3.2.3
        //reason: placement of variables in registers
        val = p_frame[ser_no][pix_no]; //get D<bit> of p_frame[ser_no][pix_no]
        val >>= bit; //interesting bit is now in D0 of val
        val <<= 7; //shift bit into D7 of val (all other bits cleared)
        byte >>= 1; //shift right bits already in byte
        byte |= val; //put bit into D7 of byte
        byte_bits++; //count bits in byte
        if( byte_bits == 8 ) //byte complete
        {
          *p_out = byte; //save byte
          p_out++; //advance pointer
          byte_bits = 0; //restart with empty byte
        }
      } //for( ser_no ...
      //fill last byte with zeros if not completed
      if( byte_bits > 0 )
      {
        byte >>= 8 - byte_bits; //fill byte with 8 - byte_bits zeros
        *p_out = byte; //save byte
        p_out++;  //advance pointer
        byte_bits = 0; //restart with empty byte
      } //if( byte_bits ...
    } //for( bit ...
    //sync bit for every serial port
    byte = pix_no == 0 ? 0xFF : 0x00; //sync bit is set for first pixel
    for( byte_bits = 0; byte_bits < td_out_ser_cnt; byte_bits += 8 ) //+= 8 because there are 8 bits in a byte
    {
      *p_out = byte;
      p_out++;
    }
  } //for( pix_no ...
}

//main program
int main( int arg_cnt, char * * args )
{
  char * p_config_file;
  unsigned int in_buffer_len, out_buffer_len;
  unsigned char * p_in_buffer;
  unsigned char * * * p_frame_buffer;
  unsigned char * * p_out_buffer;
  st_td_fmt * p_td_fmt;
  int in_sock_fd, out_sock_fd;
  struct sockaddr_in addr;
  fd_set fdset;
  struct timeval timeout;
  unsigned short dev_no, ser_no;
  int cnt, len, addr_len, i;
  unsigned char * ptr;
  in_addr_t in_addr;

  //output message
  printf( "\n"
          "TROIA distributor\n"
          "version 0.1 date 2004-08-31\n"
          "Copyright (C) 2003-2004 1stein <1stein@schuermans.info>\n"
          "Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html\n"
          "\n" );

  //no config file supplied
  if( arg_cnt < 2 )
  {
    printf( "no config file supplied (syntax: \"troia_dist <config-file>\"), using \"troia_dist.conf\"\n" );
    p_config_file = "troia_dist.conf";
  }
  //config file supplied
  else
    p_config_file = args[1];

  //read config file and get settings
  td_config_get( p_config_file );
  printf( "configuration read: %d devices with %d serial ports with %d pixels\n",
          td_out_dev_cnt, td_out_ser_cnt, td_out_pix_cnt );

  //allocate input buffer
  in_buffer_len = 65536; //maximum size of a datagram
  p_in_buffer = (unsigned char *)malloc( in_buffer_len );
  if( p_in_buffer == NULL )
  {
    printf( "could not allocate buffer (input) of size %d\n\n", in_buffer_len );
    return -1;
  }

  //allocate frame buffer
  len = td_out_dev_cnt * sizeof( unsigned char * * ) //dev_cnt pointers to 2D arrays
      + td_out_dev_cnt * td_out_ser_cnt * sizeof( unsigned char * ) //dev_cnt * ser_cnt pointers to 1D arrays
      + td_out_dev_cnt * td_out_ser_cnt * td_out_pix_cnt * sizeof( unsigned char ); //dev_cnt * ser_cnt * pix_cnt pixels
  p_frame_buffer = (unsigned char * * *)malloc( len );
  if( p_frame_buffer == NULL )
  {
    free( p_in_buffer );
    printf( "could not allocate buffer (frame) of size %d\n\n", len );
    return -1;
  }
  //initialize frame buffer array structure
  ptr = (unsigned char *)p_frame_buffer + td_out_dev_cnt * sizeof( unsigned char * * ); //array of pointers to 2D arrays
  for( dev_no = 0; dev_no < td_out_dev_cnt; dev_no++ ) //arrays of pointers to 1D arrays
  {
    p_frame_buffer[dev_no] = (unsigned char * *)ptr;
    ptr += td_out_ser_cnt * sizeof( unsigned char * );
  }
  for( dev_no = 0; dev_no < td_out_dev_cnt; dev_no++ ) //arrays of pixels
  {
    for( ser_no = 0; ser_no < td_out_ser_cnt; ser_no++ )
    {
      p_frame_buffer[dev_no][ser_no] = ptr;
      ptr += td_out_pix_cnt * sizeof( unsigned char );
    }
  }
  //clear frame buffer
  for( dev_no = 0; dev_no < td_out_dev_cnt; dev_no++ )
    for( ser_no = 0; ser_no < td_out_ser_cnt; ser_no++ )
      memset( p_frame_buffer[dev_no][ser_no], 0x00, td_out_pix_cnt * sizeof( unsigned char ) );
  
  //allocate output buffer
  out_buffer_len = ((td_out_ser_cnt + 7) & ~7) * td_out_pix_cnt; //length of output buffer for 1 device
                                                               //(round number of serial ports to next full 8)
  len = td_out_dev_cnt * sizeof( unsigned char * ) //dev_cnt pointers to 1D arrays
      + td_out_dev_cnt * out_buffer_len; //dev_cnt output buffers
  p_out_buffer = (unsigned char * *)malloc( len );
  if( p_out_buffer == NULL )
  {
    free( p_frame_buffer );
    free( p_in_buffer );
    printf( "could not allocate buffer (output) of size %d\n\n", len );
    return -1;
  }
  //initialize output buffer array structure
  ptr = (unsigned char *)p_out_buffer + td_out_dev_cnt * sizeof( unsigned char * ); //array of pointers to 1D arrays
  for( dev_no = 0; dev_no < td_out_dev_cnt; dev_no++ ) //output buffers
  {
    p_out_buffer[dev_no] = ptr;
    ptr += out_buffer_len;
  }

  //generate format from format file
  p_td_fmt = td_fmt_load( td_fmt_file );
  if( p_td_fmt == NULL )
  {
    free( p_out_buffer );
    free( p_frame_buffer );
    free( p_in_buffer );
    printf( "could not create format from \"%s\"\n\n", td_fmt_file );
    return -1;
  }
  printf( "format file read: %dx%d pixels\n", p_td_fmt->width, p_td_fmt->height );

  //create input socket
  in_sock_fd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
  if( in_sock_fd == -1 )
  {
    td_fmt_free( p_td_fmt );
    free( p_out_buffer );
    free( p_frame_buffer );
    free( p_in_buffer );
    printf( "could not create input socket: error: %s\n\n", strerror( errno ) );
    return -1;
  }

  //bind input socket
  addr.sin_family = AF_INET;
  addr.sin_port = htons( td_in_listen_port );
  addr.sin_addr.s_addr = td_in_listen_addr;
  if( bind( in_sock_fd, (struct sockaddr *)&addr, sizeof( addr ) ) == -1 )
  {
    close( in_sock_fd );
    td_fmt_free( p_td_fmt );
    free( p_out_buffer );
    free( p_frame_buffer );
    free( p_in_buffer );
    printf( "could not bind input socket to \"%d.%d.%d.%d:%d\" (udp): error: %s\n\n",
      ((unsigned char *)&addr.sin_addr.s_addr)[0],
      ((unsigned char *)&addr.sin_addr.s_addr)[1],
      ((unsigned char *)&addr.sin_addr.s_addr)[2],
      ((unsigned char *)&addr.sin_addr.s_addr)[3],
      ntohs( addr.sin_port ),
      strerror( errno ) );
    return -1;
  }
  printf( "input socket listening on \"%d.%d.%d.%d:%d\" (udp)\n",
    ((unsigned char *)&addr.sin_addr.s_addr)[0],
    ((unsigned char *)&addr.sin_addr.s_addr)[1],
    ((unsigned char *)&addr.sin_addr.s_addr)[2],
    ((unsigned char *)&addr.sin_addr.s_addr)[3],
    ntohs( addr.sin_port ) );

  //create output socket
  out_sock_fd = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP );
  if( out_sock_fd == -1 )
  {
    close( in_sock_fd );
    td_fmt_free( p_td_fmt );
    free( p_out_buffer );
    free( p_frame_buffer );
    free( p_in_buffer );
    printf( "could not create output socket: error: %s\n\n", strerror( errno ) );
    return -1;
  }

  //unblock output socket
  i = 1;
  if( ioctl( out_sock_fd, FIONBIO, &i ) == -1 )
  {
    close( out_sock_fd );
    close( in_sock_fd );
    td_fmt_free( p_td_fmt );
    free( p_out_buffer );
    free( p_frame_buffer );
    free( p_in_buffer );
    printf( "could not unblock output socket: error: %s\n\n", strerror( errno ) );
    return -1;
  }

  //bind output socket
  addr.sin_family = AF_INET;
  addr.sin_port = htons( td_out_bind_port );
  addr.sin_addr.s_addr = td_out_bind_addr;
  if( bind( out_sock_fd, (struct sockaddr *)&addr, sizeof( addr ) ) == -1 )
  {
    close( out_sock_fd );
    close( in_sock_fd );
    td_fmt_free( p_td_fmt );
    free( p_out_buffer );
    free( p_frame_buffer );
    free( p_in_buffer );
    printf( "could not bind output socket to \"%d.%d.%d.%d:%d\" (udp): error: %s\n\n",
      ((unsigned char *)&addr.sin_addr.s_addr)[0],
      ((unsigned char *)&addr.sin_addr.s_addr)[1],
      ((unsigned char *)&addr.sin_addr.s_addr)[2],
      ((unsigned char *)&addr.sin_addr.s_addr)[3],
      ntohs( addr.sin_port ),
      strerror( errno ) );
    return -1;
  }
  printf( "output socket bound to \"%d.%d.%d.%d:%d\" (udp)\n",
    ((unsigned char *)&addr.sin_addr.s_addr)[0],
    ((unsigned char *)&addr.sin_addr.s_addr)[1],
    ((unsigned char *)&addr.sin_addr.s_addr)[2],
    ((unsigned char *)&addr.sin_addr.s_addr)[3],
    ntohs( addr.sin_port ) );

  //install signal handler to end program
  signal( SIGTERM, end_signal );
  signal( SIGINT, end_signal );
  signal( SIGHUP, end_signal );

  //main loop
  while( ! end )
  {
    //wait for reception of a datagram
    FD_ZERO( &fdset );
    FD_SET( in_sock_fd, &fdset );
    timeout.tv_sec = 0;
    timeout.tv_usec = 100000;
    cnt = select( in_sock_fd + 1, &fdset, NULL, NULL, &timeout );
    if( cnt == -1 && errno != EINTR )
    {
      printf( "select returned error: %s\n\n", strerror( errno ) );
      break;
    }
    //reception of a datagram
    if( cnt > 0 && FD_ISSET( in_sock_fd, &fdset ) )
    {
      //get datagram
      addr_len = sizeof( addr );
      len = recvfrom( in_sock_fd, p_in_buffer, in_buffer_len, 0,
                      (struct sockaddr *)&addr, &addr_len );
      if( len >= 0 )
      {
        if( td_verbose )
          printf( "reception from \"%d.%d.%d.%d:%d\" (udp) (length=%d)\n",
            ((unsigned char *)&addr.sin_addr.s_addr)[0],
            ((unsigned char *)&addr.sin_addr.s_addr)[1],
            ((unsigned char *)&addr.sin_addr.s_addr)[2],
            ((unsigned char *)&addr.sin_addr.s_addr)[3],
            ntohs( addr.sin_port ),
            len );
        //processs datagram
        if( proto_datagram( p_in_buffer, (unsigned int)len, p_td_fmt, p_frame_buffer ) )
        {
          //convert frame data of every device into output format
          for( dev_no = 0; dev_no < td_out_dev_cnt; dev_no++ )
            frame2out( p_frame_buffer[dev_no], p_out_buffer[dev_no] );
          //send output data using output socket
          if( td_verbose )
            printf( "outputting data to %d devices (length=%d)\n", td_out_dev_cnt, out_buffer_len );
          in_addr = td_out_ip_base;
          for( dev_no = 0; dev_no < td_out_dev_cnt; dev_no++ )
          {
            //assemble destination address
            addr.sin_family = AF_INET;
            addr.sin_port = htons( td_out_dest_port );
            addr.sin_addr.s_addr = in_addr;
            //send
            if( sendto( out_sock_fd, p_out_buffer[dev_no], out_buffer_len, 0,
                        (struct sockaddr *)&addr, sizeof( addr ) ) != (int)out_buffer_len )
            {
              if( td_verbose )
                printf( "could not send to \"%d.%d.%d.%d:%d\" (udp) (length=%d): error: %s\n",
                  ((unsigned char *)&addr.sin_addr.s_addr)[0],
                  ((unsigned char *)&addr.sin_addr.s_addr)[1],
                  ((unsigned char *)&addr.sin_addr.s_addr)[2],
                  ((unsigned char *)&addr.sin_addr.s_addr)[3],
                  ntohs( addr.sin_port ),
                  out_buffer_len,
                  strerror( errno ) );
            }
            //calculate IP address of next device
            in_addr += td_out_ip_step;
          } //for( dec_no ...
        } //if( proto_datagram( ...
      } //if( len ...
    } // if( cnt > 0 ...
  } //while( ! end )
  printf( "terminated\n\n" );

  //clean up
  close( out_sock_fd );
  close( in_sock_fd );
  td_fmt_free( p_td_fmt );
  free( p_out_buffer );
  free( p_frame_buffer );
  free( p_in_buffer );

  return 0;
}
